// (c) UK Andrew Giddings 2003
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public Licence as published by the
// Free Software Foundation; either version 2 of the Licence, or (at your
// option) any later version.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence
// for more details.  (http://www.gnu.org/licenses/gpl.txt)

package uk.co.cix.gidds.io; 

import java.io.*;

/*******************************************************************************
* A BufferedReader which maintains the number of characters read.
*
* This is a bit rough-and-ready.  Mark/reset are not supported, because of the
* complications they cause.  Also, remember that this is counting characters,
* not bytes; the two may not match up, depending upon the default character
* encoding.  But apart from all that, it seems to work :)
*
* @author Andrew Giddings
*******************************************************************************/

public class PositionReader
	extends java.io.BufferedReader
{
	protected long charsRead = 0;

	/**
	* Create a buffering character-input stream that uses an input buffer of
	* the specified size.
	*
	* @param in A Reader
	* @param sz Input-buffer size
	*
	* @exception IllegalArgumentException If sz is <= 0
	**/
	public PositionReader(Reader in, int sz)
	{
		super(in, sz);
	}

	/**
	* Create a buffering character-input stream that uses a default-sized
	* input buffer.
	*
	* @param in A Reader
	**/
	public PositionReader(Reader in)
	{
		super(in);
	}

	/**
	* Read a single character.
	* As {@link java.io.BufferedReader#read()}.
	**/
	public int read()
		throws IOException
	{
		int result = super.read();
		if (result >= 0)
			++charsRead;
		return result;
	}

	/**
	* Read characters into a portion of an array.
	* As {@link java.io.BufferedReader#read(char[], int, int)}.
	**/
	public int read(char cbuf[], int off, int len)
		throws IOException
	{
		int result = super.read(cbuf, off, len);
		charsRead += result;
		return result;
	}

	/**
	* Read a line of text.
	* As {@link java.io.BufferedReader#readLine()}.
	**/
	public String readLine()
		throws IOException
	{
		String result = super.readLine();
		if (result != null)
			charsRead += result.length() + 1;
		return result;
	}

	/**
	* Skip characters.
	* As {@link java.io.BufferedReader#skip(long)}.
	**/
	public long skip(long n)
		throws IOException
	{
		long result = super.skip(n);
		charsRead += result;
		return result;
	}

	/**
	* Tell whether this stream supports the mark() operation, which it doesn't.
	* (BufferedReader does, but my patience has limits.)
	**/
	public boolean markSupported()
	{
		return false;
	}

	public void mark(int readAheadLimit)
	{
		throw new RuntimeException("Mark/reset not supported for this stream");
	}

	public void reset()
	{
		throw new RuntimeException("Mark/reset not supported for this stream");
	}

	public long getCharsRead()
	{
		return charsRead;
	}
}
