// (c) UK Andrew Giddings 2003
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public Licence as published by the
// Free Software Foundation; either version 2 of the Licence, or (at your
// option) any later version.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence
// for more details.  (http://www.gnu.org/licenses/gpl.txt)

package uk.co.cix.gidds.misc;

import java.awt.*;
import java.io.*;
import java.util.Vector;
import java.awt.event.KeyEvent;

import uk.co.cix.gidds.io.StreamFilter;
import uk.co.cix.gidds.io.TextAreaWriter;
import uk.co.cix.gidds.io.PositionReader;
import uk.co.cix.gidds.ui.StdDialog;

/*******************************************************************************
* A wrapper for a {@link StreamFilter} text-processing engine that exposes it
* as a full application.
*
* @author Andrew Giddings
*******************************************************************************/

public class FileAppWrapper
{
	public static final String BACKUP_EXTENSION = ".bak";

	private static String filename;
	private static Vector options;
	private static long length;
	private static PositionReader reader;

	/**
	* Called from the main() method, this invokes the stream filter on any files
	* supplied; if none, then brings up a file dialog for file selection.  The
	* file is replaced by the processed version; the original is renamed to .bak
	* first.
	*
	* @param filter The stream filter (text processor) to use.
	* @param title The app's title.
	* @param args The command-line arguments.  Any with a leading '-'
	*             are parsed as flags.
	**/
	public static void wrapReplacing(StreamFilter filter, String title, String[] args)
		throws Exception
	{
		options = new Vector();
		boolean hadAnyFiles = false;

		for (int arg = 0; arg < args.length; ++arg)
		{	if (args[arg].startsWith("-"))
				options.add(args[arg].substring(1));
			else
			{	hadAnyFiles = true;
				filename = args[arg];
				processReplacing(filter);
			}
		}

		// If no filenames were specified, then show a file dialog
		// and process the files chosen.
		if (!hadAnyFiles)
		{	FileDialog fd = new FileDialog(StdDialog.getDummyFrame(title),
								title, FileDialog.LOAD);
			fd.pack();

			for (;;)
			{	fd.show();
				if (fd.getFile() == null)
					break;	// Done.
				filename = fd.getDirectory() + fd.getFile();
				try
				{	processReplacing(filter);
					StdDialog.okDialog(null, title,
									filename + " processed successfully");
				}
				catch (Exception e)
				{	e.printStackTrace();
					StdDialog.cancelDialog(null, title,
								new String[] {"Can't process " + filename,
												e.toString()});
				}
			}
		}
	}

	private static String[] toStringArray(Vector v)
	{
		String[] array = new String[v.size()];
		for (int i = v.size() - 1; i >= 0; --i)
			array[i] = (String) v.elementAt(i);
		return array;
	}

	private static void processReplacing(StreamFilter filter)
		throws Exception
	{
		File toFile = new File(filename);
		length = toFile.length();
		File fromFile = new File(filename + BACKUP_EXTENSION);

		if (fromFile.exists())
			fromFile.delete();
		if (!toFile.renameTo(fromFile))
			throw new IOException("Can't rename " + toFile.getPath()
										+ " to " + fromFile.getPath());

		reader = new PositionReader(new FileReader(fromFile));
		try
		{	BufferedWriter writer = new BufferedWriter(new FileWriter(toFile));
			try
			{	filter.filterStream(reader, writer);
			}
			finally
			{	writer.close();
			}
		}
		finally
		{	reader.close();
		}
	}

	/**
	* Called from the main() method, this invokes the stream filter
	* on any filenames supplied, writing the results to stdout.
	* If '-' is specified on its own, then reads stdin.
	* If no files are specified, uses file dialogs to select file(s) to
	* process, and writes the results in a window.
	*
	* @param filter The stream filter (text processor) to use.
	* @param title The app's title.
	* @param args The command-line arguments.  Options with a leading
	* '-' are recognised and passed to the filter.
	* @param gfx
	**/
	public static void wrapToStdout(StreamFilter filter, String title, String[] args)
		throws Exception
	{
		options = new Vector();
		boolean hadAnyFiles = false;

		BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(System.out));

		try
		{	for (int arg = 0; arg < args.length; ++arg)
			{	if (args[arg].startsWith("-") && args[arg].length() > 1)
					options.add(args[arg].substring(1));
				else
				{	hadAnyFiles = true;
					filename = args[arg];
					if (args[arg].equals("-"))
					{	reader = new PositionReader(new InputStreamReader(System.in));
						length = -1;
					}
					else
					{	reader = new PositionReader(new FileReader(args[arg]));
						length = new File(args[arg]).length();
					}
					try
					{
						filter.filterStream(reader, writer);
					}
					finally
					{	reader.close();
					}
				}
			}

			// If no filenames were specified, then show a file dialog
			// and process the files chosen.
			if (!hadAnyFiles)
			{	FileDialog fd = new FileDialog(StdDialog.getDummyFrame(title),
								title, FileDialog.LOAD);

				final StdDialog d = new StdDialog(null, title,
						new String[] {"Continue", "Quit"},
						new int[] {'\n', KeyEvent.VK_ESCAPE});
				TextArea textArea = new TextArea("", 8, 60,
						TextArea.SCROLLBARS_VERTICAL_ONLY);
				d.add(textArea, BorderLayout.CENTER);
				d.pack();
				BufferedWriter textWriter = new BufferedWriter(new TextAreaWriter(textArea));

				for (;;)
				{	fd.show();
					if (fd.getFile() == null)
						break;	// Done.

					filename = fd.getDirectory() + fd.getFile();
					length = new File(filename).length();
					textArea.setText("");
					try
					{	reader = new PositionReader(new FileReader(filename));
						filter.filterStream(reader, textWriter);
						reader.close();
						textWriter.flush();
						d.show();
					}
					catch (Exception e)
					{	e.printStackTrace();
						StdDialog.cancelDialog(null, title,
									new String[] {"Can't process " + filename,
													e.toString()});
					}
				}
			}
		}
		finally
		{	writer.close();
		}
	}

	/**
	* @return The name of the current file being processed, or "-" for stdin.
	**/
	public static String getFilename()
	{
		return filename;
	}

	/**
	* @return an array of any options that were included on the command line
	* (excluding leading hyphens).
	**/
	public static String[] getOptions()
	{
		return toStringArray(options);
	}

	/**
	* @return The size of the current file, in bytes,
	* or -1 if reading from stdin.
	**/
	public static long getFileSize()
	{
		return length;
	}

	/**
	* @return How far we are through the current file, as a percentage.
	* <code>null</code> if we're reading from stdin, so can't tell.
	**/
	public static String getFilePercentageDone()
	{
		if (filename.equals("-"))
			return null;

		long percentage = reader.getCharsRead() * 100 / length;
		return new Long(percentage).toString() + "%";
	}
}
