// (c) UK Andrew Giddings 2004
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public Licence as published by the
// Free Software Foundation; either version 2 of the Licence, or (at your
// option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence
// for more details.  (http://www.gnu.org/licenses/gpl.txt)

package uk.co.cix.gidds.text;

import java.util.*;
import java.io.*;
import java.awt.GridLayout;
import javax.swing.*;

import uk.co.cix.gidds.io.StreamFilter;
import uk.co.cix.gidds.io.IOUtils;
import uk.co.cix.gidds.misc.FileAppWrapper;
import uk.co.cix.gidds.ui.JOptionFrame;

/*******************************************************************************
* App to convert American English spelling to British English, as nature
* intended!
*
* This uses a {@link StringDFA Deterministic Finite-state Automaton (DFA)} to
* do the matching, so is impressively fast (after the initial set-up).<p>
*
* The list of replacements to make is stored in a file, {@link #SPELLING_FILE}
* - see its comments for details on the format.  The DFA is built from this
* file each time, so changes to the file will take effect next run.<p>
*
* Normally, most replacements will be made silently; a few (such as tire->tyre,
* meter->meter) depend on context, and so the user will be asked about these.
* With the <code>-a</code> flag, the user is asked to confirm all
* replacements; with <code>-n</code>, none.<p>
*
* Due to DFA size limitations, only Unicode chars 0-255 can be matched; other
* characters are ignored in matching (though can appear in replacements).<p>
*
* TO DO:
* De/serialise DFA where possible instead of creating from fresh each time?
* Parameterise by replacements file etc. to allow other sorts of replacement?
* (Split out diacritics, common mistakes, hyphenation, etc.?)
* All/None should operate on complete words, not just replacement substrings.
* If the replacements file has a comment by any questionable cases, show that
* comment when asking the user?
* '!' matches letters >256 as well as non-letters...
*
* @author Andrew Giddings
* @version 1.01 2004-07-04
*******************************************************************************/

public class Anglicise
	implements StreamFilter
{
	public static void main(String[] args)
		throws Exception
	{
		System.out.println(TITLE + "  " + COPYRIGHT);
		System.out.println(LICENCE);
		FileAppWrapper.wrapReplacing(new Anglicise(), "Anglicise", args);
		System.exit(0);
	}

	private static final String TITLE = "ANGLICISE v1.01";
	private static final String COPYRIGHT = "(c) UK  Andrew Giddings  2004";
	private static final String LICENCE = "Released under the GPL (see http://www.gnu.org/licenses/gpl.txt).";

	public static final String SPELLING_FILE = "uk/co/cix/gidds/text/Anglicise_replacements.txt";

	private StringDFA _dfa = null;
	private int _mode = ASK_QUESTIONABLES;
	private boolean _quitting = false;
	private JOptionFrame _frame;
	private JLabel _label1, _label2;

	// Enumeration: replacement types
	private static final int ALWAYS = 0;
	private static final int QUESTIONABLE = 1;
	private static final int NORMAL = 2;
	private static final int NEVER = 3;

	// Enumeration: run modes
	private static final int ASK_ALL = 0;
	private static final int ASK_QUESTIONABLES = 1;
	private static final int ASK_NONE = 2;

	private static final int CONTEXT_SIZE = 45;
	private static final String LABEL_PREFIX = "<html><font size=\"+1\">";

	// Button shortcodes
	private static final int YES_MNEMONIC = 'y';
	private static final int NO_MNEMONIC = 'n';
	private static final int ALL_MNEMONIC = 'a';
	private static final int NONE_MNEMONIC = 'o';

	/***************************************************************************
	* Represents a generic replacement to perform.  Stored in the DFA.
	***************************************************************************/
	private static class Replacement
		implements Serializable
	{
		public final String match, replaceString;
		public int when;

		public Replacement(String match, String replaceString, int when)
		{	this.match = match;
			this.replaceString = replaceString;
			this.when = when;
		}
	}

	/**
	* Read a stream of text and perform replacements on it.
	* This is the main body of the app, called by {@link FileAppWrapper}.
	**/
	public void filterStream(BufferedReader in, BufferedWriter out)
		throws Exception
	{
		String[] options = FileAppWrapper.getOptions();
		if (options.length > 0)
		{	if (options.length == 1 && options[0].equals("a"))
				_mode = ASK_ALL;
			if (options.length == 1 && options[0].equals("n"))
				_mode = ASK_NONE;
			else
			{	System.err.println("Flags: -a to ask about all replacements, -n to ask about none.");
				System.exit(0);
			}
		}

		initDFA();

		// Scan through the file, replacing as we go
		String prevLine, line = "", nextLine = in.readLine();
		for (;;)
		{
			prevLine = line;
			line = nextLine;
			if (line == null)
				break;
			nextLine = in.readLine();

			line += IOUtils.LINE_SEPARATOR;	// So we can recognise all words at EOL
			_dfa.reset();
			_dfa.nextState(' ');	// So we can recognise all words at SOL

			for (int i = 0; i < line.length(); ++i)
			{	char c = line.charAt(i);
				Replacement replacement = (Replacement) _dfa.nextState((c < 256) ? c : ' ');
				if (replacement != null)
				{
					int start = i - replacement.match.length() + 1;
					if (replacement.match.startsWith("!"))
						++start;
					int end = i;
					if (replacement.match.endsWith("!"))
						--end;

					String replaceString = TextUtils.copyCase(
											replacement.replaceString,
											line.substring(start, end + 1));

					if (okayToReplace(prevLine, line, nextLine, start, end, replaceString, replacement))
					{
						line = line.substring(0, start) + replaceString + line.substring(end + 1);

						// Back to the start of the word
						i += replacement.replaceString.length() - (end + 1 - start);
						do
						{	--i;
						} while (i > -1 && Character.isLetter(line.charAt(i)));
						_dfa.reset();
						_dfa.nextState(' ');	// So we can recognise all words at SOL
					}
				}
			}

			out.write(line);
		}
	}

	/**
	* Set up the DFA.  Read a list of spelling corrections from a text file,
	* set up an array of character folds, and construct a corresponding DFA.
	* (This is relatively time-consuming, hence only done once per run.  N.B.
	* reusing the DFA also reuses any always/never replacement user choices...)
	**/
	private void initDFA()
		throws IOException
	{
		if (_dfa != null)
			return;		// Already done

		List matches = new ArrayList();
		List targets = new ArrayList();
		InputStream is = ClassLoader.getSystemResourceAsStream(SPELLING_FILE);
		if (is == null)
			throw new FileNotFoundException("Spelling file " + SPELLING_FILE + " not in classpath");
		BufferedReader spellfile = new BufferedReader(new InputStreamReader(is));
		try
		{
			int lineNo = 0;
			String line;
			while ((line = spellfile.readLine()) != null)
			{	++lineNo;

				// Strip comments
				int i = line.indexOf('#');
				if (i >= 0)
				{	while (i > 0 && Character.isWhitespace(line.charAt(i - 1)))
						--i;
					line = line.substring(0, i);
				}

				// Strip blank lines
				if (line.length() == 0)
					continue;

				// Valid replacement?
				i = line.indexOf('\t');
				if (i < 0)
					throw new IllegalArgumentException("Only one field on line " + lineNo);
				String match = line.substring(0, i);

				if (line.indexOf('\t', i + 1) >= 0)
					throw new IllegalArgumentException("More than two fields on line " + lineNo);
				String replace = line.substring(i + 1);

				if (replace.indexOf(match) >= 0)
					throw new IllegalArgumentException("Recursive replacement ('" + match + "' by '" + replace + "')");

				int when = NORMAL;
				if (replace.endsWith("?"))
				{	when = QUESTIONABLE;
					replace = replace.substring(0, replace.length() - 1);
				}

				matches.add(match);
				targets.add(new Replacement(match, replace, when));
			}
		}
		finally
		{
			spellfile.close();
		}

		// Set up folds array
		char[] folds = new char[256];
		final String LETTERS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
		for (int i = 0; i < LETTERS.length(); ++i)
		{	char c = LETTERS.charAt(i);
			folds[c] = c;
			folds[Character.toLowerCase(c)] = c;
		}
		folds['-'] = '-';

		_dfa = new StringDFA((String[]) matches.toArray(new String[] {}),
							targets.toArray(new Replacement[] {}), folds);
	}

	private void initFrame()
	{
		if (_frame == null)
		{	_label1 = new JLabel((String) null, JLabel.CENTER);
			_label2 = new JLabel((String) null, JLabel.CENTER);
			JPanel panel = new JPanel(new GridLayout(2, 1));
			panel.add(_label1);
			panel.add(_label2);
			_frame = new JOptionFrame("", panel,
				new String[] {"Yes", "No", "All", "None"},
				new int[] {YES_MNEMONIC, NO_MNEMONIC, ALL_MNEMONIC, NONE_MNEMONIC}, 1);
		}

		String percentage = FileAppWrapper.getFilePercentageDone();
		if (percentage == null)
			_frame.setTitle("Anglicise " + FileAppWrapper.getFilename());
		else
			_frame.setTitle("Anglicise " + FileAppWrapper.getFilename() + " (" + percentage + ")");
	}

	/**
	* Check whether we want to perform a particular replacement.
	* Some replacements are always done; some never; and for some we must ask
	* the user.
	**/
	private boolean okayToReplace(String prevLine, String line, String nextLine,
							int start, int end,
							String replace, Replacement replacement)
	{
		if (_quitting || replacement.when == NEVER)
			return false;
		if (replacement.when == ALWAYS)
			return true;
		if (_mode == ASK_NONE)
			return replacement.when == NORMAL;
		if (_mode == ASK_QUESTIONABLES && replacement.when == NORMAL)
			return true;

		// Okay, so we need to ask the user.

		// Expand the original and replacement to a full word for display
		int i = start;
		while (i > 0 && Character.isLetter(line.charAt(i - 1)))
			--i;
		replace = line.substring(i, start) + replace;
		start = i;

		i = end;
		while (i + 1 < line.length() && Character.isLetter(line.charAt(i + 1)))
			++i;
		replace += line.substring(end + 1, i + 1);
		end = i;

		// And clip a suitable context to show
		if (prevLine.endsWith(IOUtils.LINE_SEPARATOR))
			prevLine = prevLine.substring(0, prevLine.length() - 1);
		if (line.endsWith(IOUtils.LINE_SEPARATOR))
			line = line.substring(0, line.length() - 1);
		if (nextLine == null)
			nextLine = "";
		else if (nextLine.endsWith(IOUtils.LINE_SEPARATOR))
			nextLine = nextLine.substring(0, nextLine.length() - 1);
		String context = prevLine + " " + line.substring(0, start)
					+ "<b>" + line.substring(start, end + 1)
					+ "</b>" + line.substring(end + 1) + " " + nextLine;

		int cStart = start + prevLine.length() + 1;
		int cEnd = end + prevLine.length() + 1;
		if (context.length() > cEnd + CONTEXT_SIZE)
		{	i = context.indexOf(" ", cEnd + CONTEXT_SIZE);
			if (i > 0)
				context = context.substring(0, i);
		}
		if (cStart > CONTEXT_SIZE)
		{	i = context.lastIndexOf(" ", cStart - CONTEXT_SIZE);
			if (i > 0)
			{	context = context.substring(i + 1);
				cStart -= i + 1;
				cEnd -= i + 1;
			}
		}

		initFrame();
		_label1.setText(LABEL_PREFIX + "Replace <b>"
						+ line.substring(start, end + 1) + "</b> with <b>"
						+ replace + "</b> in:");
		_label2.setText(LABEL_PREFIX + context);
		int result = _frame.getResponse();
//		_label1.setText(null);
//		_label2.setText(null);
// For some reason, these HTML labels take *ages* to display... and hide!
// (Plain text ones would be better, but we need to highlight the change...)

		if (result == YES_MNEMONIC)
			return true;

		if (result == ALL_MNEMONIC)
		{	replacement.when = ALWAYS;
			return true;
		}

		if (result == NONE_MNEMONIC)
			replacement.when = NEVER;
		else if (result == JOptionFrame.WINDOW_CLOSED)
			_quitting = true;

		return false;
	}
}
