// (c) UK Andrew Giddings 2004
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public Licence as published by the
// Free Software Foundation; either version 2 of the Licence, or (at your
// option) any later version.         
//
// This program is distributed in the hope that it will be useful, but        
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence 
// for more details.  (http://www.gnu.org/licenses/gpl.txt)   

package uk.co.cix.gidds.text;

import java.io.*;

/*******************************************************************************
* A Deterministic Finite-state Automaton, used for matching multiple fixed
* strings.<p>
*
* It defines a set of states; some of these signify that a string has been
* matched, and have an associated `target' object which can hold anything you
* like.  It also defines the transitions between these states, each of which is
* determined by a single character.  To search some text, we start at the root
* state, and then read characters one by one, moving from state to state
* accordingly.  Whenever we reach a state with a target, we've just matched a
* string.<p>
* 
* The only restrictions are:<ul>
* <li>as only one target is returned at each state
* transition, one string can't equal or end with another.
* It copes perfectly when one string /begins/ with another, though, or has
* another as a proper substring, and finds all the overlapping matches.
* <li>max 32768 nodes, for space reasons.
* </ul>(Neither is inherent to the algorithm; both could be removed with some
* recoding.)<p>
*
* To enable case-insensitive matching etc., characters may be `folded' before
* comparison.  Reducing the number of distinct characters to match on in this
* way will also reduce the size of the DFA proportionately.<p>
*
* Once created, the DFA may be serialised to avoid the creation overhead next
* time.  However, this will also serialise all of the target objects, which may
* not be the Right Thing.  (If it is, you may want to {@link #optimise()}
* first.)  When deserialised, as when created, the DFA is reset to the root
* state, ready to begin matching.<p>
*
* The algorithm for creating the DFA from the list of strings to match is my
* own.  Although there's plenty of literature on creating
* <i>Non</i>-deterministic FAs, simulating them would be much too slow given
* the number of strings we could be searching for.  I couldn't find any
* relevant literature on setting up a DFA, so I've devised the method here.
* It's relatively slow (order M * N, where N is the number of strings to match,
* and M is the total number of characters in those strings), but seems to be
* fast enough for general use; and after that set-up each state transition
* takes only a couple of array lookups, regardless of the number of strings to
* be matched.
*
* @author Andrew Giddings
* @version 1.01 2004-07-09
*******************************************************************************/

public class StringDFA
	implements java.io.Serializable
{
	private char[] _folds;
	private short[][] _transits;	// Sparse array, with sub-arrays only for used characters
	private Object[] _targets;
	private short _nextFreeState;

	private transient short _currentState = 0;

	/**
	* Construct a fold array initialised so that every character `folds' to
	* itself.  This array is not shared and may be safely modified.
	*
	* @param size The size of the array to return.
	**/
	public static char[] getIdentityFold(int size)
	{
		char[] identityFold = new char[size];
		for (char c = 0; c < size; ++c)
			identityFold[c] = c;
		return identityFold;
	}

	/**
	* Create a DFA to match the given strings.  The string itself is returned
	* when a match is found.
	* Characters are not folded prior to comparison.  (But only single-byte
	* characters are accepted.)
	*
	* @param strings The strings to search for.
	*
	* @throws IllegalArgumentException if one string ends with, or is a
	*                                  duplicate of, another.
	**/
	public StringDFA(String[] strings)
	{
		init(strings, strings, getIdentityFold(256));
	}

	/**
	* Create a DFA to match the given strings, returning the corresponding
	* target object when a match is found.
	* Characters are not folded prior to comparison.  (But only single-byte
	* characters are accepted.)
	*
	* @param strings The strings to search for.
	*
	* @param targets Arbitrary objects to be returned when each of the strings
	*                is matched.
	*
	* @throws IllegalArgumentException if one string ends with, or is a
	*                                  duplicate of, another.
	**/
	public StringDFA(String[] strings, Object[] targets)
	{
		init(strings, targets, getIdentityFold(256));
	}

	/**
	* Create a DFA to match the given strings, each returning the corresponding
	* target object.
	*
	* @param strings The strings to search for.
	*
	* @param targets Arbitrary objects to be returned when each of the strings
	*                is matched.
	*
	* @param folds An array specifying how characters are `folded' prior to
	*              comparison.
	*
	* @throws IllegalArgumentException if one string ends with, or is a
	*                                  duplicate of, another.
	**/
	public StringDFA(String[] strings, Object[] targets, char[] folds)
	{
		init(strings, targets, folds);
	}

	/**
	* Create the DFA.
	*
	* @throws IllegalArgumentException if one string ends with, or is a
	* duplicate of, another.
	**/
	private void init(String[] strings, Object[] targets, char[] folds)
	{
		if (strings.length != targets.length)
			throw new IllegalArgumentException("Must be as many strings ("
										+ strings.length + ") as targets ("
										+ targets.length + ")");
		_folds = folds;

		// Get an upper bound for the number of states needed.
		// (Worst case is 1 state for each letter in each search string, plus 1
		// for the root state.  That's only attained if every search string
		// starts with a different letter, but it's rarely more than double.)
		short states = 1;
		for (int i = 0; i < strings.length; ++i)
			states += strings[i].length();

		// Get the number of character codes needed after folding
		short maxFoldedChar = 0;
		for (int i = 0; i < _folds.length; ++i)
			if (maxFoldedChar < _folds[i])
				maxFoldedChar = (short) _folds[i];

		// Create the arrays.
		// The _transits[] array is a sparse two-dimensional array.  The first
		// dimension is by the character determining the transition; entries
		// for characters never matched are set to null.  The remainder hold
		// arrays indexed by the state transitioning from; these will be
		// created as needed.
		// Note that as 0 is the root state, all states will link to the root
		// state by default for all transitions.  Which is nice.
		_transits = new short[maxFoldedChar + 1][];
		_targets = new Object[states];
		_nextFreeState = 1;

		// Add all strings, starting from the root.
		// This creates a partial machine which will match any string at the
		// start of the text.  It forms a nice simple tree structure.
		for (int i = 0; i < strings.length; ++i)
			addFromRoot(strings[i], targets[i]);

		// And set interconnects for overlapping words and resets.
		// This is the tricky bit, which allows strings to match starting from
		// any state, not just the root state.
		// (Note that the tree structure above gives all children greater state
		// numbers than their parents; so here the outer loop by state number
		// will visit all parents before their children.  This ensures we handle
		// abc/bcd/cde overlapping cases no matter what order they're presented
		// in.)
		for (short n = 1; n < _nextFreeState; ++n)
			for (int i = 0; i < strings.length; ++i)
			{	char c = _folds[strings[i].charAt(0)];
				short topState = _transits[c][0];
				if (_transits[c][n] == 0)
					_transits[c][n] = topState;
				else if (_transits[c][n] != topState)
					addLinked(n, strings[i], targets[i]);
			}
	}

	/**
	* Add a new string, starting from the root state.
	* When we run out of existing states, this creates fresh ones.
	* It also creates the states arrays as needed for each new character.
	*
	* @param target An object to flag a successful match.
	*
	* @throws IllegalArgumentException if the word has already been added.
	**/
	private void addFromRoot(String s, Object target)
	{
		 short state = 0, len = (short) s.length();

		for (int i = 0; i < len; ++i)
		{	char c = _folds[s.charAt(i)];
			if (_transits[c] == null)
				_transits[c] = new short[_targets.length];
			short newState = _transits[c][state];
			if (newState == 0)
			{	newState = _nextFreeState++;
				_transits[c][state] = newState;
			}
			state = newState;
		}

		if (_targets[state] != null)
			throw new IllegalArgumentException("Duplicate string: `" + s + "'");

		_targets[state] = target;
	}

	/**
	* Set the transitions so we can match an overlapping string.
	* When we run out of existing states, instead of creating new ones, this
	* simply links to the state we'd have got to if we'd recognised this string
	* straight from the root state.  Alternatively, if this word is a substring
	* of an existing one, it adds the target onto an existing state.  This is
	* where we spot words that end with others, as there'll be an existing
	* target there.
	*
	* @param state A state from which the string might be recognised
	*
	* @throws IllegalArgumentException if this string turns out to be the end
	* of another
	**/
	private void addLinked(short state, String s, Object target)
	{
		short linkedState = 0, len = (short) s.length();
		for (int i = 0; i < len; ++i)
		{	char c = _folds[s.charAt(i)];
			linkedState = _transits[c][linkedState];
			short newState = _transits[c][state];
			if (newState == 0)
			{
				// Set the link
				_transits[c][state] = linkedState;
				return;
			}
			else if (newState == linkedState)
				// Already done
				return;

			state = newState;
		}

		// Is this already the end of a different string?
		// N.B. We may already have marked it as the end of /this/ one (e.g.
		// case for strings `for' `the' and `other'...) so allow that.
		if (_targets[state] != null && _targets[state] != target)
			throw new IllegalArgumentException("String `" + s + "' is the end of another");

		// No, so we've found a proper substring.  So add a target for that too.
		// Note that this is the only circumstance in which we'll get two
		// different states with the same target.  But that's right, as we still
		// need to distinguish whether we're in the middle of matching the
		// longer string or not.
		_targets[state] = target;
	}

	/**
	* Reduce the sizes of internal arrays.
	* We probably use fewer than the max number of nodes we estimated during
	* construction, as some strings probably have common initial letters.  So
	* copy to new arrays to save space.
	* This may save e.g. 40% on a DFA with 400 words (less on smaller ones).
	* Subsequent calls are ignored.
	**/
	public void optimise()
	{
		if (_nextFreeState == _targets.length)
			return;		// Already optimised

		Object[] newTargets = new Object[_nextFreeState];
		System.arraycopy(_targets, 0, newTargets, 0, newTargets.length);
		_targets = newTargets;
		for (int i = 0; i < _transits.length; ++i)
		{	if (_transits[i] != null)
			{	short[] newTransits = new short[_nextFreeState];
				System.arraycopy(_transits[i], 0, newTransits, 0, newTransits.length);
				_transits[i] = newTransits;
			}
		}
	}

	/**
	* Write the DFA to stdout in a long-winded, human-readable fashion.
	* For debugging purposes.
	**/
	public void display()
	{
		StringBuffer sb = new StringBuffer();
		for (int i = 0; i < _nextFreeState; ++i)
		{
			sb.setLength(0);
			sb.append(Integer.toString(i));
			sb.append(":");
			if (_targets[i] != null)
			{
				sb.append(" [");
				sb.append(_targets[i].toString());
				sb.append(']');
			}
			for (int j = 0; j < _folds.length; ++j)
				if (_transits[j] != null && _transits[j][i] != 0)
				{
					sb.append(' ');
					sb.append((char) j);
					sb.append(':');
					sb.append(_transits[j][i]);
				}
			System.out.println(sb.toString());
		}
		System.out.println(_nextFreeState + " nodes used out of a max " + _targets.length);
	}

	/**
	* Make the transition from the current state to that indicated by the
	* given character.
	*
	* @return a target object, if this completes a match.  (Further overlapping
	* matches may follow - call {@link #reset} to start from scratch.)
	* Otherwise, returns <code>null</code>.
	**/
	public Object nextState(char c)
	{
		if (c < 0 || c >= _folds.length)
			throw new IllegalArgumentException("Invalid character: char `" + c + "' + (" + ((int) c) + ") >= " + _folds.length);

		short[] t = _transits[_folds[c]];
		if (t == null)
			_currentState = 0;
		else
			_currentState = t[_currentState];
		return _targets[_currentState];
	}

	/**
	* Reset the DFA to the root state, ready to start matching from scratch.
	**/
	public void reset()
	{
		_currentState = 0;
	}

	/**
	* For testing...
	* Create a DFA to match words given in the arguments, then use it to match
	* stdin.  Stops at end-of-file (Ctrl+D).
	**/
	public static void main(String[] args)
		throws java.io.IOException
	{
		java.util.Date start = new java.util.Date();
		StringDFA d = new StringDFA(args);
		java.util.Date end = new java.util.Date();
		d.display();
		System.out.println("Creation took " + (end.getTime() - start.getTime()) + " ms.");
		System.out.println("-----");

		ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream("serial"));
		oos.writeObject(d);
		oos.close();

		d.optimise();
		oos = new ObjectOutputStream(new FileOutputStream("serialO"));
		oos.writeObject(d);
		oos.close();

		int c = 0;
		for (;;)
		{	int i = System.in.read();
			if (i < 0)	// End of file (i.e. Ctrl+D)
				break;
			if (i == '\n')
				c = 0;
			else
				++c;
			Object o = d.nextState((char) i);
			if (o != null)
				System.out.println("Char " + c + ": " + o);
		}
	}
}
