// (c) UK Andrew Giddings 2003
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public Licence as published by the
// Free Software Foundation; either version 2 of the Licence, or (at your
// option) any later version.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence
// for more details.  (http://www.gnu.org/licenses/gpl.txt)

package uk.co.cix.gidds.text;

import java.util.*;

/*******************************************************************************
* Various text-related utility functions.
*******************************************************************************/

public final class TextUtils
{
	/**
	* Convert a string into a sequence of bytes.
	* (This truncates each char to 8 bits.)
	**/
	private static byte[] getBytes(String s)
	{
		int l = s.length();
		char[] chars = new char[l];
		s.getChars(0, l, chars, 0);
		byte[] bytes = new byte[l];
		for (int i = 0; i < l; ++i)
			bytes[i] = (byte) chars[i];
		// That's long-winded, but about 5 times faster than the equivalent
		// byte[] bytes = s.getBytes("UTF-16BE");

		return bytes;
	}

	/**
	* Convert a string into a sequence of bytes.
	* (This truncates each char to 8 bits.)
	**/
	private static byte[] getBytes(StringBuffer s)
	{
		int l = s.length();
		char[] chars = new char[l];
		s.getChars(0, l, chars, 0);
		byte[] bytes = new byte[l];
		for (int i = 0; i < l; ++i)
			bytes[i] = (byte) chars[i];
		// That's long-winded, but about 5 times faster than the equivalent
		// byte[] bytes = s.getBytes("UTF-16BE");

		return bytes;
	}

	/**
	* Convert a string from the platform's default encoding into Unicode.
	* Most of the file I/O classes perform this conversion automatically;
	* this should be used in the cases it's not (e.g. with RandomAccessFile).
	* N.B. Assumes input chars are 0-FF; high bits will be lost.
	**/
	public static String convertToUnicode(String s)
	{
		return new String(getBytes(s));
	}

	/**
	* As {@link #convertToUnicode(String)}.
	**/
	public static String convertToUnicode(StringBuffer s)
	{
		return new String(getBytes(s));
	}

	/**
	* Convert a string from the given encoding into Unicode.
	* N.B. Assumes input chars are 0-FF; high bits will be lost.
	**/
	public static String convertToUnicode(String s, String encoding)
	{
		try
		{	return new String(getBytes(s), encoding);
		}
		catch (java.io.UnsupportedEncodingException e)
		{	e.printStackTrace();
			return s;
		}
	}

	/**
	* @return Whether the string consists entirely of digits.
	**/
	public static boolean isDigits(String s)
	{
		for (int i = 0; i < s.length(); ++i)
			if (!Character.isDigit(s.charAt(i)))
				return false;
		return true;
	}

	/**
	* Convenience method to replace a certain number of characters inside a
	* given string with another.
	**/
	public static String replace(String s, int start, int length, String replace)
	{
		return s.substring(0, start) + replace + s.substring(start + length);
	}

	/**
	* Convenience method to insert a given string inside another.
	**/
	public static String insert(String s, int pos, String insert)
	{
		return s.substring(0, pos) + insert + s.substring(pos);
	}

	/**
	* Convenience method to remove a certain number of characters from a string.
	**/
	public static String delete(String s, int pos, int length)
	{
		return s.substring(0, pos) + s.substring(pos + length);
	}

	/**
	* Convenience method to replace every occurrence of one substring with
	* another.
	**/
	public static String replaceAll(String s, String search, String replace)
	{	StringBuffer sb = new StringBuffer(s);
		int i;
		while ((i = sb.indexOf(search)) >= 0)
			sb.replace(i, i + search.length(), replace);
		return sb.toString();
	}

	/**
	* Adjust a string to use the same letter case as another.
	* (For example, adjust a text replacement to use the same case as what it's
	* replacing.)
	*
	* The two strings will usually start and end with the same letters, so use
	* those directly.  For the middle bit that's different, use the same case
	* character-by-character, repeating the case of the last character if the
	* replacement's longer.  (Since most words have the same case after the
	* first couple of letters, this simple method should suffice.)
	**/
	public static void copyCase(StringBuffer s, String example)
	{
		// Set case for a matching prefix
		int f = 0, rs = s.length() - 1, re = example.length() - 1;
		while (f <= rs && f <= re
		&& Character.toLowerCase(s.charAt(f))
		== Character.toLowerCase(example.charAt(f)))
		{	s.setCharAt(f, example.charAt(f));
			++f;
		}

		// Set case for a matching suffix
		while (re >= f && rs >= f
		&& Character.toLowerCase(s.charAt(rs))
		== Character.toLowerCase(example.charAt(re)))
			s.setCharAt(rs--, example.charAt(re--));

		// Set case for the rest (if any).
		if (f <= rs)
		{
			char c = example.charAt(Math.min(f, example.length() - 1));
			for (int i = f; i <= rs; ++i)
			{
				if (i <= re)
					c = example.charAt(i);

				if (Character.isTitleCase(c))
					s.setCharAt(i, Character.toTitleCase(s.charAt(i)));
				else if (Character.isUpperCase(c))
					s.setCharAt(i, Character.toUpperCase(s.charAt(i)));
				else
					s.setCharAt(i, Character.toLowerCase(s.charAt(i)));
			}
		}
	}

	/**
	* Adjust a string to use the same letter case as another.
	* See {@link #copyCase(StringBuffer, String)}.
	**/
	public static String copyCase(String s, String example)
	{
		StringBuffer result = new StringBuffer(s);
		copyCase(result, example);
		return result.toString();
	}

	private static Vector v = new Vector();

	/**
	* Break up a string into an array of words.
	* (Not thread-safe.)
	**/     
	public static String[] splitWords(String s)
	{
		StringTokenizer st = new StringTokenizer(s);
		while (st.hasMoreTokens())
			v.addElement(st.nextToken());
		String[] array = new String[v.size()];
		v.copyInto(array);
		v.removeAllElements();
		return array;
	}

	/**
	* Break up a string into an array of hashcodes of words.
	* (Not thread-safe.)
	**/
	public static long[] splitWordsToHashCodes(String s)
	{
		StringTokenizer st = new StringTokenizer(s);
		while (st.hasMoreTokens())
			v.addElement(st.nextToken());
		long[] array = new long[v.size()];
		for (int i = 0; i < array.length; ++i)
			array[i] = v.elementAt(i).hashCode();
		v.removeAllElements();
		return array;
	}

	/**
	* Check whether a character is digit-like (i.e. a digit or a symbol
	* involving digits).
	**/
	public static boolean isDigitLike(char c)
	{
		return Character.isDigit(c)
			|| c == '' || c == '' || c == ''
			|| c == '' || c == '' || c == '';
	}

	/**
	* Extra a word from a given string immediately before a given position
	* (which is assumed not to be within a word).
	* null if none.
	**/
	public static String getPrevWord(String s, int pos)
	{
		int last = pos - 1;
		while (last >= 0 && !Character.isLetter(s.charAt(last)))
			--last;
		if (last < 0)
			return null;
		int first = last - 1;
		while (first >= 0 && (Character.isLetter(s.charAt(first)) || s.charAt(first) == '\'' || s.charAt(first) == ''))
			--first;
		return s.substring(first + 1, last + 1);
	}

	/**
	* Extra the next word from a given string immediately after a given position
	* (which is assumed not to be within a word).
	* null if none.
	**/
	public static String getNextWord(String s, int pos)
	{
		int length = s.length();
		int first = pos + 1;
		while (first < length && !Character.isLetter(s.charAt(first)))
			++first;
		if (first >= length)
			return null;
		if (first > 0 && (s.charAt(first - 1) == '\'' || s.charAt(first - 1) == ''))
			--first;
		int last = first + 1;
		while (last < length && (Character.isLetter(s.charAt(last)) || s.charAt(last) == '\'' || s.charAt(last) == ''))
			++last;
		if (last > 0 && (s.charAt(last - 1) == '\'' || s.charAt(last - 1) == ''))
			--last;
		return s.substring(first, last);
	}
}
