// (c) UK Andrew Giddings 2003
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public Licence as published by the
// Free Software Foundation; either version 2 of the Licence, or (at your
// option) any later version.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence
// for more details.  (http://www.gnu.org/licenses/gpl.txt)

package uk.co.cix.gidds.ui;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/*******************************************************************************
* A ready-rolled frame with a message and some buttons, rather like {@link
* JOptionPane}.  But this is a frame rather than a dialog, so it doesn't vanish
* when the user responds.<p>
*
* (One reason for this is to work around an apprent bug in Mac OS X's JDK 1.4.1,
* where showing too many dialogs in succession sometimes freezes the GUI.  But
* it's also substantially quicker, and more manageable for the user.)<p>
*
* It also allows keyboard mnemonics for the buttons (with or without the usual
* modifier key), and Escape to close the window.
* And it expands the window to its preferred size each time (though without
* contracting the window afterwards).<p>
*
* At times when we're NOT awaiting a user response, the hourglass cursor is
* used, and clicks within the frame are ignored.  (Though the window can still
* be closed, which gets reported next time we await a response.)
*
* @author Andrew Giddings
*******************************************************************************/

public class JOptionFrame
{
	public static final int WINDOW_CLOSED = 0;
	private static final int NO_RESPONSE_YET = -9999;

	private final JFrame frame;
	private final JButton[] buttons;
	private final int[] mnemonics;

	private volatile int response = NO_RESPONSE_YET;
	private boolean disposed = false;

	/**
	* Create the frame.
	*
	* @param title The frame title.
	* @param body The main body of the frame, i.e. everything except the
	*             buttons.  Its appearance will usually be changed each time a
	*             response is required.
	* @param buttonTitles The titles of the buttons to use, in order.
	* @param buttonMnemonics The mnemonic codes for the buttons.  These should
	*             all be distinct, and different from {@link #WINDOW_CLOSED}.
	*             If any of the corresponding keys are pressed, the appropriate
	*             button will be selected.
	* @param defaultButton The index of the button to use as the default one.
	**/
	public JOptionFrame(String title, JComponent body, String[] buttonTitles,
						int[] buttonMnemonics, int defaultButton)
	{
		if (buttonTitles.length != buttonMnemonics.length)
			throw new IllegalArgumentException(buttonTitles.length
					+ " buttons != " + buttonMnemonics.length + " mnemonics!");

		frame = new JFrame(title);
		frame.getContentPane().setLayout(new BorderLayout());
		frame.getContentPane().add(body, BorderLayout.CENTER);
		JPanel panel = new JPanel(new FlowLayout(FlowLayout.TRAILING, 10, 10));
		frame.getContentPane().add(panel, BorderLayout.SOUTH);
		buttons = new JButton[buttonTitles.length];
		mnemonics = buttonMnemonics;

		KeyListener kl = new KeyAdapter()
			{	public void keyTyped(KeyEvent e)
				{	for (int i = 0; i < mnemonics.length; ++i)
						if (e.getKeyChar() == mnemonics[i])
							buttons[i].doClick();
				}
				public void keyPressed(KeyEvent e)
				{	if (e.getKeyCode() == 27)	// Escape key
						JOptionFrame.this.userResponded(WINDOW_CLOSED);
				}
			};

		for (int i = 0; i < buttonTitles.length; ++i)
		{	buttons[i] = new JButton(buttonTitles[i]);
			final int mnemonic = mnemonics[i];
			buttons[i].setMnemonic(mnemonic);
			buttons[i].addActionListener(new ActionListener()
				{	public void actionPerformed(ActionEvent e)
					{	JOptionFrame.this.userResponded(mnemonic);
					}
				});
			buttons[i].addKeyListener(kl);
			panel.add(buttons[i]);
		}

		frame.getRootPane().setDefaultButton(buttons[defaultButton]);

		frame.setDefaultCloseOperation(frame.DO_NOTHING_ON_CLOSE);
		frame.addWindowListener(new WindowAdapter()
			{	public void windowClosing(WindowEvent e)
				{	JOptionFrame.this.userResponded(WINDOW_CLOSED);
				}
			});

		frame.pack();
		UIUtils.centreOnScreen(frame);
		JUIUtils.initHourglass(frame);
		JUIUtils.setHourglass(frame, true);
	}

	/**
	* Called when the user has made some sort of response.
	**/
	private synchronized void userResponded(int response)
	{	this.response = response;
		notify();
		if (response == WINDOW_CLOSED)
			dispose();
	}

	/**
	* Get a response from the user.  This method blocks until the user makes a
	* response by clicking one of the buttons, pressing an appropriate key, or
	* closing the window.
	*
	* @return The mnemonic of the button or key pressed, or {@link #WINDOW_CLOSED}.
	**/
	public synchronized int getResponse()
	{
		if (disposed)
			return WINDOW_CLOSED;

		UIUtils.expandToPreferredSize(frame, 50, 20);
		JUIUtils.setHourglass(frame, false);

		if (!frame.isShowing())
			frame.show();

		// (Don't reset the response variable here, so we can recognise
		// window-close events since the last response.)

		while (response == NO_RESPONSE_YET)
		{	try
			{	wait();
			}
			catch (InterruptedException e)
			{	// Ignore
			}
		}

		JUIUtils.setHourglass(frame, true);

		int retval = response;
		response = NO_RESPONSE_YET;
		return retval;
	}

	/**
	* Hide and dispose of the frame. After this, calls to {@link #getResponse()}
	* will immediately return {@link #WINDOW_CLOSED}.
	**/
	public void dispose()
	{
		frame.hide();
		frame.dispose();
		disposed = true;
	}

	public void setTitle(String title)
	{
		frame.setTitle(title);
	}
}
