// (c) UK Andrew Giddings 2003
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public Licence as published by the
// Free Software Foundation; either version 2 of the Licence, or (at your
// option) any later version.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence
// for more details.  (http://www.gnu.org/licenses/gpl.txt)

package uk.co.cix.gidds.ui;

import java.awt.*;
import java.awt.event.*;

/*******************************************************************************
* A Dialog with standard features.  Designed mainly to mimic the system
* dialogs on EPOC, but should be generic enough.
*
* It has a row of buttons along the bottom, not FocusTraversible, but
* which respond when the appropriate keys are pressed, wherever the
* focus is. (Escape is always recognised.)
* It also allows the use of up and down arrows to move between fields,
* and Tab to drop down Choices.
* TextFields initially have all text selected.
* There are many static convenience methods for standard dialogs.
*
* N.B.  Dialogs aren't dispose()d when finished!  This is deliberate, as on EPOC
* dispose()ing the dialog seems to redeliver the keypress to the next window...
*
* @author Andrew Giddings
*******************************************************************************/

public class StdDialog
	extends CentredDialog
	implements KeyListener
{
	private int key;
	private int keys[];
//	private Button buttons[];

	private static final Font BUTTON_FONT = new Font("SansSerif",
														Font.BOLD, 11);

	private static Frame dummyFrame;

	public static Frame getDummyFrame(String title)
	{
		if (dummyFrame == null)
		{	dummyFrame = new Frame(title);
//			dummyFrame.pack();
			dummyFrame.show();
			dummyFrame.hide();
		}
		else
			dummyFrame.setTitle(title);
		return dummyFrame;
	}

	public StdDialog(Frame frame, String title, String[] labels, int[] keys)
	{
		super(frame == null ? getDummyFrame(title) : frame, title, true);
		this.keys = keys;

		if (labels.length != keys.length)
		{	dispose();
			throw new RuntimeException("Label/key mismatch");
		}

//		buttons = new Button[keys.length];
		Panel bp = new Panel(new FlowLayout(FlowLayout.CENTER, 10, 5));
		add(bp, BorderLayout.SOUTH);
		for (int i = 0; i < labels.length; ++i)
		{	Button b = new Button(labels[i])
				{	public boolean isFocusTraversable()
					{	return false;
					}
				};
			b.setFont(BUTTON_FONT);
			b.setActionCommand(String.valueOf(keys[i]));
			bp.add(b);
//			buttons[i] = b;
			b.addActionListener(new ActionListener()
				{	public void actionPerformed(ActionEvent e)
					{	key = Integer.parseInt(((Button)
										e.getSource()).getActionCommand());
						StdDialog.this.setVisible(false);
//						StdDialog.this.dispose();
					}
				});
		}
	}

	public void setVisible(boolean visible)
	{
		if (visible)
			addListeners(this);
		super.setVisible(visible);
	}

	// Add the appropriate listeners to every (non-Container) component
	// within this Container (recursively).  This will include the buttons.
	private void addListeners(Container con)
	{
		Component[] c = con.getComponents();
		for (int i = 0; i < c.length; ++i)
			if (c[i] instanceof Container)
				addListeners((Container) c[i]);
			else
			{	c[i].addKeyListener(this);
				if (c[i] instanceof TextField)
					((TextField) c[i]).selectAll();
			}
	}

	public int getKey()
	{
		return key;
	}

	private void handleKey(KeyEvent e, int key)
	{
		for (int i = keys.length - 1; i >= 0; --i)
			if (key == keys[i] || key == KeyEvent.VK_ESCAPE)
			{	e.consume();
// This /should/ simulate a button press.  Shame it doesn't work on EPOC :(
//				buttons[i].dispatchEvent(new KeyEvent(buttons[i],
//				KeyEvent.KEY_PRESSED, System.currentTimeMillis(),
//				0, KeyEvent.VK_ENTER, (char) KeyEvent.VK_ENTER));

				this.key = key;
				setVisible(false);

// If we dispose(), then on EPOC the KeyEvent seems to be reapplied to the
// parent frame, despite consume(), setKeyChar() or whatever :(
//				dispose();
				return;
			}
	}

	// implements KeyListener
	public void keyPressed(KeyEvent e)
	{
		int key = e.getKeyCode();
		if (key == KeyEvent.VK_DOWN
		&& !(e.getSource() instanceof TextArea))
		{	e.setKeyCode(KeyEvent.VK_TAB);
			e.setModifiers(0);
		}
		else if (key == KeyEvent.VK_UP
		&& !(e.getSource() instanceof TextArea))
		{	e.setKeyCode(KeyEvent.VK_TAB);
			e.setModifiers(InputEvent.SHIFT_MASK);
		}
		else if (key == KeyEvent.VK_TAB
		&& e.getSource() instanceof Choice)
			e.setModifiers(InputEvent.CTRL_MASK);
		else
			handleKey(e, key);
	}

	// implements KeyListener
	public void keyReleased(KeyEvent e)
	{
	}

	// implements KeyListener
	public void keyTyped(KeyEvent e)
	{
	}

	// Convenience methods

	private static final String[] OK_LABELS = new String[] {"OK"};
	private static final String[] CANCEL_LABELS = new String[] {"Cancel"};
	private static final String[] OK_CANCEL_LABELS
									= new String[] {"Cancel", "OK"};
	private static final int[] OK_KEYS = new int[] {'\n'};
	private static final int[] CANCEL_KEYS = new int[] {KeyEvent.VK_ESCAPE};
	private static final int[] OK_CANCEL_KEYS
									= new int[] {KeyEvent.VK_ESCAPE, '\n'};

	public static void okDialog(Frame frame, String title, String line)
	{
		standardDialog(frame, title, new String[] {line}, OK_LABELS, OK_KEYS);
	}

	public static void okDialog(Frame frame, String title, String[] lines)
	{
		standardDialog(frame, title, lines, OK_LABELS, OK_KEYS);
	}

	public static void cancelDialog(Frame frame, String title, String line)
	{
		standardDialog(frame, title, new String[] {line},
						CANCEL_LABELS, CANCEL_KEYS);
	}

	public static void cancelDialog(Frame frame, String title, String[] lines)
	{
		standardDialog(frame, title, lines, CANCEL_LABELS, CANCEL_KEYS);
	}

	public static boolean okCancelDialog(Frame frame, String title,
											String line)
	{
		return standardDialog(frame, title, new String[] {line},
						OK_CANCEL_LABELS, OK_CANCEL_KEYS) == '\n';
	}

	public static boolean okCancelDialog(Frame frame, String title,
											String[] lines)
	{
		return standardDialog(frame, title, lines, OK_CANCEL_LABELS,
						OK_CANCEL_KEYS) == '\n';
	}

	public static boolean doCancelDialog(Frame frame, String title,
											String line, String doLabel)
	{
		return standardDialog(frame, title, new String[] {line},
						new String[] {"Cancel", doLabel},
						OK_CANCEL_KEYS) == '\n';
	}

	public static boolean doCancelDialog(Frame frame, String title,
											String[] lines, String doLabel)
	{
		return standardDialog(frame, title, lines,
						new String[] {"Cancel", doLabel},
						OK_CANCEL_KEYS) == '\n';
	}

	public static int standardDialog(Frame frame, String title,
									String[] lines, String[] labels, int[] keys)
	{
		Component[] c = new Component[lines.length];
		for (int i = 0; i < lines.length; ++i)
			c[i] = new Label(lines[i], Label.CENTER);

		return standardDialog(frame, title, c, labels, keys);
	}

	public static int standardDialog(Frame frame, String title,
									Component[] lines, String[] labels, int[] keys)
	{
		StdDialog d = new StdDialog(frame, title, labels, keys);

		Panel p = new Panel(new GridLayout(0, 1, 4, 4));
		for (int i = 0; i < lines.length; ++i)
			p.add(lines[i]);
		d.add(p, BorderLayout.CENTER, 0);
			// (0 ensures the panel is first in the dialogue & gets focus)
		d.pack();

		// Now space things out a bit.
		// If only GridLayout let you set the insets or something...
		Dimension mySize = d.getSize();
		mySize.width += 16;
		mySize.height += 16;
		d.setSize(mySize);
		d.setVisible(true);
		return d.getKey();
	}
}
